<?php

/**
 * Class Query
 */
class Query {

    const SELECT = 'select';
    const INSERT = 'insert';
    const UPDATE = 'update';
    const DELETE = 'delete';

    /**
     * @var string
     */
    protected $action;
    /**
     * @var string
     */
    protected $table;
    /**
     * @var array
     */
    protected $select;
    /**
     * @var array
     */
    protected $from;
    /**
     * @var array
     */
    protected $innerJoin;
    /**
     * @var array
     */
    protected $leftJoin;
    /**
     * @var array
     */
    protected $where;
    /**
     * @var array
     */
    protected $order;
    /**
     * @var array
     */
    protected $group;
    /**
     * @var array
     */
    protected $having;
    /**
     * @var int
     */
    protected $limit;
    /**
     * @var int
     */
    protected $offset;
    /**
     * @var array
     */
    protected $values;
    /**
     * @var bool
     */
    protected $hasAlias;
    /**
     * @var bool
     */
    protected $ignore;

    /**
     * Constructor
     */
    public function __construct()
    {
        $this->reset();
    }

    public function reset()
    {
        $this->select = array();
        $this->from = array();
        $this->innerJoin = array();
        $this->leftJoin = array();
        $this->where = array();
        $this->order = array();
        $this->group = array();
        $this->having = array();
        $this->limit = null;
        $this->offset = null;
        $this->table = null;
        $this->action = null;
        $this->values = array();
        $this->hasAlias = true;
        $this->ignore = false;

        return $this;
    }

    /**
     * @return $this
     *
     * if empty select, set by default '*'
     */
    public function select()
    {
        $this->setAction(self::SELECT);

        $args = func_get_args();
        if(!empty($args)) {
            $this->addInProperty('select', $args);
        }

        return $this;
    }

    /**
     * @param $table
     * @return $this
     */
    public function insert($table)
    {
        $this->setAction(self::INSERT);
        $this->table = $table;

        return $this;
    }

    /**
     * @param $table
     * @return $this
     */
    public function update($table)
    {
        $this->setAction(self::UPDATE);
        $this->table = $table;

        return $this;
    }

    /**
     * @param $table
     * @return $this
     */
    public function delete($table)
    {
        $this->setAction(self::DELETE);
        $this->table = $table;

        return $this;
    }

    /**
     * @return $this
     *
     * list of the table
     */
    public function from()
    {
        $args = func_get_args();
        if(!empty($args)) {
            $this->addInProperty('from', $args);
        }

        return $this;
    }

    /**
     * @return $this
     *
     * Example :
     *  - $q->innerJoin('table1', 't.my_field', 'y.field_2');
     *      => generate : INNER JOIN table1 t ON(t.my_field = y.field_2)
     *
     *  - $q->innerJoin('table1', 't.my_field', '!=', 'y.field_2');
     *      => generate : INNER JOIN table1 t ON(t.my_field != y.field_2)
     *
     *  - $q->innerJoin('table1', 'OR', array('y.field_1', 'tx.field_2'), array('tx.ff_2', 'y.ff'));
     *      => generate : INNER JOIN table1 t ON(y.field_1 = tx.field_2 OR tx.ff_2 = y.ff)
     *
     *  - $q->innerJoin('table1', 'AND', array('y.field_1', 'tx.field_2'), array('tx.ff_2', 'y.ff'));
     *      => generate : INNER JOIN table1 t ON(y.field_1 = tx.field_2 AND tx.ff_2 = y.ff)
     *
     */
    public function innerJoin()
    {
        $this->join('innerJoin', func_get_args());
        return $this;
    }

    /**
     * @return $this
     *
     * see innerJoin, replace name function
     */
    public function leftJoin()
    {
        $this->join('leftJoin', func_get_args());
        return $this;
    }

    /**
     * @return $this
     *
     * Example :
     *  - $q->where('al.my_field = "myvalue"');
     *  - $q->where("my_field", "myvalue"); equivalent to $q->where("my_field", "=", "myvalue");
     *  - $q->where("my_field", "<>", "myvalue");
     */
    public function where()
    {
        $nbArgs = func_num_args();

        if($nbArgs < 1 || $nbArgs > 3) {
            throw new BadMethodCallException('Invalid number argument : 2 or 3 arguments');
        }

        $args = func_get_args();

        if ($nbArgs == 1) {
            $this->where[] = $args[0];
        } else {
            if ($nbArgs == 2) {
                list($field, $value) = $args;
                $operator = ($value == "null") ? 'IS' : '=';
            } else {
                list($field, $operator, $value) = $args;
            }
            if (is_string($value) && $value != "null") {
                $value = '"' . $value . '"';
            }
            $this->where[] = $field . ' ' . $operator . ' ' . $value;
        }

        return $this;
    }

    /**
     * @return $this
     */
    public function order()
    {
        $args = func_get_args();
        if(!empty($args)) {
            $this->addInProperty('order', $args);
        }

        return $this;
    }

    /**
     * @return $this
     */
    public function group()
    {
        $args = func_get_args();
        if(!empty($args)) {
            $this->addInProperty('group', $args);
        }

        return $this;
    }

    /**
     * @return $this
     */
    public function having()
    {
        $nbArgs = func_num_args();

        if($nbArgs < 1 || $nbArgs > 3) {
            throw new BadMethodCallException('Invalid number argument : 2 or 3 arguments');
        }

        $args = func_get_args();

        if ($nbArgs == 1) {
            $this->having[] = $args[0];
        } else {
            if ($nbArgs == 2) {
                list($field, $value) = $args;
                $operator = '=';
            } else {
                list($field, $operator, $value) = $args;
            }
            if (is_string($value)) {
                $value = '"' . $value . '"';
            }
            $this->having[] = $field . ' ' . $operator . ' ' . $value;
        }

        return $this;
    }

    /**
     * @param $value
     * @return $this
     */
    public function limit($value)
    {
        $this->limit = $value;
        return $this;
    }

    /**
     * @param $value
     * @return $this
     */
    public function offset($value)
    {
        $this->offset = $value;
        return $this;
    }

    /**
     * @param $field
     * @param $value
     * @return $this
     */
    public function value($field, $value)
    {
        $this->values[$field] = $value;
        return $this;
    }

    /**
     * @param bool $bool
     * @return $this
     */
    public function setHaveAlias($bool)
    {
        $this->hasAlias = (bool)$bool;
        return $this;
    }

    /**
     * @param bool $bool
     * @return $this
     */
    public function setIgnore($bool)
    {
        $this->ignore = (bool)$bool;
        return $this;
    }

    /**
     * @return string
     */
    public function build()
    {
        if(empty($this->action)) {
            throw new BadMethodCallException('no action defined');
        }

        $sql = '';

        if($this->action == self::SELECT) {
            if (empty($this->select)) {
                $this->select[] = '*';
            }
            $sql .= "SELECT " . implode(',', $this->select);
        }

        if($this->action == self::INSERT) {
            $sql .= "INSERT".($this->ignore ? " IGNORE": "")." INTO";
        }

        if($this->action == self::UPDATE) {
            $sql .= "UPDATE";
        }

        if($this->action == self::DELETE) {
            $sql .= "DELETE FROM";
        }

        if($this->action == self::SELECT) {
            $tables = array();
            foreach ($this->from as $f) {
                $tables[] = $f .($this->hasAlias ? ' ' . self::getAlias($f) : '');
            }
            $sql .= " FROM (" . implode(',', $tables) . ")";


            // INNER JOIN
            if (!empty($this->innerJoin)) {
                $sql .= ' INNER JOIN ' . implode(' INNER JOIN ', $this->innerJoin);
            }

            // LEFT JOIN
            if (!empty($this->leftJoin)) {
                $sql .= ' LEFT JOIN ' . implode(' LEFT JOIN ', $this->leftJoin);
            }
        }

        if($this->action == self::INSERT || $this->action == self::UPDATE || $this->action == self::DELETE) {
            $sql .= " " . $this->table;
        }

        if($this->action == self::INSERT || $this->action == self::UPDATE) {
            $sql.= " SET";

            $i = 0;
            foreach ($this->values as $key => $value) {
                Tools::escapeDataToSql($value);

                if($i > 0){
                    $sql.= ',';
                }
                if(in_array($value, array('NOW()', 'CURDATE()', 'NULL')) || is_null($value)) {
                    $sql.=" " . $key." = " . $value . "";
                }else{
                    $sql.=" " . $key." = '" . $value . "'";
                }
                $i++;
            }
        }

            // WHERE
        if(!empty($this->where)) {
            $sql .= " WHERE " . implode(' AND ', $this->where);
        }

        if($this->action == self::SELECT) {
            // GROUP
            if (!empty($this->group)) {
                $sql .= " GROUP BY " . implode(',', $this->group);
            }

            // HAVING
            if (!empty($this->having)) {
                $sql .= " HAVING " . implode(' AND ', $this->having);
            }

            // ORDER
            if (!empty($this->order)) {
                $sql .= " ORDER BY " . implode(',', $this->order);
            }

            // LIMIT et OFFSET
            if (!is_null($this->limit)) {
                $sql .= " LIMIT " . $this->limit;

                if (!is_null($this->offset)) {
                    $sql .= " OFFSET " . $this->offset;
                }
            }
        }

        return $sql;
    }

    /**
     * @param      $type : innerJoin | leftJoin
     * @param      $args
     * @param bool $get
     *
     * @return string|null
     */
    public function join($type, $args, $get = false)
    {
        $depth = $this->array_depth($args);

        /*if($depth > 2) {
            throw new Exception('not valid parameters');
        }*/

        if (count($args) == 1) {
            array_push($this->$type, $args[0]);
        } elseif ($depth == 2) {
            list($table, $operator) = $args;

            $arr = array();
            for ($i = 2; $i < count($args); $i++) {
                $arr = array_merge($arr, array($args[$i]));
            }

            $joins = array();
            foreach ($arr as $a) {
                array_unshift($a, $table);
                $joins[] = $this->join($type, $a, true);
            }

            $operator = ' ' . $operator . ' ';

            array_push($this->$type, $table . ' ' . self::getAlias($table) . ' ON (' . implode($operator, $joins) . ')');
        } else {
            if (count($args) == 3) {
                $operator = '=';
                list($table, $field1, $field2) = $args;
            } else {
                list($table, $field1, $operator, $field2) = $args;
            }

            if ($get) {
                return $field1 . ' ' . $operator . ' ' . $field2;
            }

            array_push($this->$type, $table . ' ' . self::getAlias($table) . ' ON (' . $field1 . ' ' . $operator . ' ' . $field2 . ')');
        }

        return null;
    }

    /**
     * @param $property
     * @param $args
     */
    private function addInProperty($property, $args)
    {
        $this->$property = array_merge($this->$property, $args);
    }

    /**
     * @param $table
     * @return string
     *
     * return an alias for table pass in parameter
     * Example :
     *  - tableName => alias t
     *  - my_table => alias mt
     */

    static public function getAlias($table)
    {
        $alias = '';
        if (strpos($table, ' ') === false) {
            $partTable = explode('_', $table);
            foreach ($partTable as $pt) {
                $alias .= substr($pt, 0, 1);
            }
        }

        return $alias;
    }

    /**
     * @param $formvars
     * @return $this
     */
    public function extras($formvars)
    {
        if(!empty($formvars)) {
            foreach ($formvars as $k => $v) {
                if (method_exists($this, $k)) {
                    $this->$k($v);
                } else {
                    $this->where($k, $v);
                }
            }
        }

        return $this;
    }

    /**
     * @param array $array
     * @return int
     *
     * return the depth of the array pass in parameter
     */
    private function array_depth(array $array) {
        $max_depth = 1;

        foreach ($array as $value) {
            if (is_array($value)) {
                $depth = $this->array_depth($value) + 1;

                if ($depth > $max_depth) {
                    $max_depth = $depth;
                }
            }
        }

        return $max_depth;
    }

    /**
     * @param string $action
     */
    private function setAction(string $action): void
    {
        if ($this->action !== null && $this->action !== $action) {
            throw new BadMethodCallException(sprintf("Action '%s' unauthorized, query already defined with action %s", $action, $this->action));
        }
        $this->action = $action;
    }
}
